"""
Current report shared resources
"""

# pylint: disable=abstract-method,arguments-renamed

# stdlib
import asyncio as aio
from datetime import date
from typing import List, Optional, Tuple, Union

# module
from avwx.base import AVWXBase
from avwx.service import get_service, NOAA_ADDS
from avwx.static.core import NA_UNITS, WX_TRANSLATIONS
from avwx.structs import Code, ReportData, ReportTrans, Units


def wx_code(code: str) -> Union[Code, str]:
    """
    Translates weather codes into readable strings

    Returns translated string of variable length
    """
    if not code:
        return ""
    ret, code_copy = "", code
    if code[0] == "+":
        ret = "Heavy "
        code = code[1:]
    elif code[0] == "-":
        ret = "Light "
        code = code[1:]
    # Return code if code is not a code, ex R03/03002V03
    if len(code) not in [2, 4, 6] or code.isdigit():
        return code
    while code:
        try:
            ret += WX_TRANSLATIONS[code[:2]] + " "
        except KeyError:
            ret += code[:2]
        code = code[2:]
    return Code(code_copy, ret.strip())


def get_wx_codes(codes: List[str]) -> Tuple[List[str], List[Code]]:
    """
    Separates parsed WX codes
    """
    other: List[str] = []
    ret: List[Code] = []
    for item in codes:
        code = wx_code(item)
        if isinstance(code, Code):
            ret.append(code)
        else:
            other.append(code)
    return other, ret


class Report(AVWXBase):
    """
    Base report to take care of service assignment and station info
    """

    #: ReportTrans dataclass of translation strings from data. Parsed on update()
    translations: Optional[ReportTrans] = None

    def __init__(self, icao: str):
        super().__init__(icao)
        if self.station is not None:
            service = get_service(icao, self.station.country)
            self.service = service(self.__class__.__name__.lower())  # type: ignore


class Reports(AVWXBase):
    """
    Base class containing multiple reports
    """

    raw: Optional[List[str]] = None  # type: ignore
    data: Optional[List[ReportData]] = None  # type: ignore
    units: Units = Units(**NA_UNITS)

    def __init__(self, icao: str = None, lat: float = None, lon: float = None):
        if icao:
            super().__init__(icao)
            if self.station is not None:
                lat = self.station.latitude
                lon = self.station.longitude
        elif lat is None or lon is None:
            raise ValueError("No station or valid coordinates given")
        self.lat = lat
        self.lon = lon
        self.service = NOAA_ADDS("aircraftreport")

    def __repr__(self) -> str:
        return f"<avwx.{self.__class__.__name__} lat={self.lat} lon={self.lon}>"

    @staticmethod
    def _report_filter(reports: List[str]) -> List[str]:
        """Applies any report filtering before updating raw_reports"""
        return reports

    async def _update(  # type: ignore
        self, reports: List[str], issued: Optional[date], disable_post: bool
    ) -> bool:
        if not reports:
            return False
        reports = self._report_filter(reports)
        return await super()._update(reports, issued, disable_post)

    def parse(self, reports: Union[str, List[str]], issued: Optional[date] = None):
        """Updates report data by parsing a given report

        Can accept a report issue date if not a recent report string
        """
        self.source = None
        if isinstance(reports, str):
            reports = [reports]
        return aio.run(self._update(reports, issued, False))

    async def async_parse(
        self, reports: Union[str, List[str]], issued: Optional[date] = None
    ):
        """Async updates report data by parsing a given report

        Can accept a report issue date if not a recent report string
        """
        self.source = None
        if isinstance(reports, str):
            reports = [reports]
        return await self._update(reports, issued, False)

    def update(
        self,
        timeout: int = 10,
        disable_post: bool = False,
    ) -> bool:
        """Updates report data by fetching and parsing the report

        Returns True if new reports are available, else False
        """
        reports = self.service.fetch(lat=self.lat, lon=self.lon, timeout=timeout)  # type: ignore
        self.source = self.service.root
        return aio.run(self._update(reports, None, disable_post))

    async def async_update(self, timeout: int = 10, disable_post: bool = False) -> bool:
        """Async updates report data by fetching and parsing the report"""
        reports = await self.service.async_fetch(  # type: ignore
            lat=self.lat, lon=self.lon, timeout=timeout
        )
        self.source = self.service.root
        return await self._update(reports, None, disable_post)



"""
Parsing for NOAA NBM forecasts
"""

# Reference: https://www.weather.gov/mdl/nbm_textcard_v32

# pylint: disable=too-many-arguments

# stdlib
from contextlib import suppress
from typing import Callable, Dict, List, Optional, Tuple

# module
from avwx import structs
from avwx.service.files import NOAA_NBM
from avwx.static.gfs import UNITS
from .base import (
    Forecast,
    _decimal_10,
    _decimal_100,
    _direction,
    _find_time_periods,
    _init_parse,
    _numbers,
    _number_100,
    _parse_lines,
    _split_line,
)

UNITS = {
    **UNITS,
    "accumulation": "in",
    "duration": "hour",
    "solar_radiation": "W/m2",
    "wave_height": "ft",
}


_CEILING = {
    # 888 is code but processed as 88800 due to postfix
    "88800": (None, "unlimited")
}
_WIND = {"NG": (0, "zero")}


def _ceiling(line: str, size: int = 3) -> List[Optional[structs.Number]]:
    """Parse line into Number objects handling ceiling special units"""
    return _numbers(line, size, postfix="00", special=_CEILING)


def _wind(line: str, size: int = 3) -> List[Optional[structs.Number]]:
    """Parse line into Number objects handling wind special units"""
    return _numbers(line, size, special=_WIND)


_HANDLERS: Dict[str, Tuple[str, Callable]] = {
    "TMP": ("temperature", _numbers),
    "DPT": ("dewpoint", _numbers),
    "SKY": ("sky_cover", _numbers),
    "WDR": ("wind_direction", _direction),
    "WSP": ("wind_speed", _wind),
    "GST": ("wind_gust", _wind),
    "DUR": ("precip_duration", _numbers),
    "PZR": ("freezing_precip", _numbers),
    "PSN": ("snow", _numbers),
    "PPL": ("sleet", _numbers),
    "PRA": ("rain", _numbers),
    "SLV": ("snow_level", _number_100),
    "SOL": ("solar_radiation", _numbers),
    "SWH": ("wave_height", _numbers),
}

_HOUR_HANDLERS: Dict[str, Tuple[str, Callable]] = {
    "P": ("precip_chance", _numbers),
    "Q": ("precip_amount", _decimal_100),
    "T": ("thunderstorm", _numbers),
    "S": ("snow_amount", _decimal_10),
    "I": ("icing_amount", _decimal_100),
}

_NBHS_HANDLERS: Dict[str, Tuple[str, Callable]] = {
    "CIG": ("ceiling", _ceiling),
    "VIS": ("visibility", _decimal_10),
    "LCB": ("cloud_base", _number_100),
    "MHT": ("mixing_height", _number_100),
    "TWD": ("transport_wind_direction", _direction),
    "TWS": ("transport_wind_speed", _numbers),
    "HID": ("haines", _numbers),
}


def _parse_factory(
    data_class,
    period_class,
    handlers: Dict[str, Tuple[str, Callable]],
    hours: int = 2,
    size: int = 3,
    prefix: int = 4,
) -> Callable:
    """Creates handler function for static and computed keys"""

    handlers = {**_HANDLERS, **handlers}

    def handle(key: str) -> Tuple:
        """Returns response key(s) and value handler for a line key"""
        with suppress(KeyError):
            return handlers[key]
        if not key[1:].isdigit():
            raise KeyError
        root, handler = _HOUR_HANDLERS[key[0]]
        return f"{root}_{key[1:].lstrip('0')}", handler

    def parse(report: str):
        """Parser for NBM reports"""
        if not report:
            return None
        data, lines = _init_parse(report)
        period_strings = _split_line(lines[hours], size, prefix)
        timestamp = data.time.dt if data.time else None
        periods = _find_time_periods(period_strings, timestamp)
        data_lines = lines[hours + 1 :]
        # Normalize line prefix length
        if prefix != 4:
            indexes = (4, prefix)
            start, end = min(indexes), max(indexes)
            data_lines = [l[:start] + l[end:] for l in data_lines]
        _parse_lines(periods, data_lines, handle, size)
        return data_class(
            raw=data.raw,
            sanitized=data.sanitized,
            station=data.station,
            time=data.time,
            remarks=data.remarks,
            forecast=[period_class(**p) for p in periods],
        )

    return parse


parse_nbh: Callable[[str], structs.NbhData] = _parse_factory(
    structs.NbhData, structs.NbhPeriod, _NBHS_HANDLERS, hours=1
)
parse_nbs: Callable[[str], structs.NbsData] = _parse_factory(
    structs.NbsData, structs.NbsPeriod, _NBHS_HANDLERS
)
parse_nbe: Callable[[str], structs.NbeData] = _parse_factory(
    structs.NbeData, structs.NbePeriod, {}, size=4, prefix=5
)


class _Nbm(Forecast):
    units = structs.NbmUnits(**UNITS)
    _service_class = NOAA_NBM  # type: ignore
    _parser: staticmethod

    async def _post_update(self):
        self.data = self._parser(self.raw)

    def _post_parse(self):
        self.data = self._parser(self.raw)


class Nbh(_Nbm):
    """Class to handle NBM NBH report data"""

    report_type = "nbh"
    _parser = staticmethod(parse_nbh)


class Nbs(_Nbm):
    """Class to handle NBM NBS report data"""

    report_type = "nbs"
    _parser = staticmethod(parse_nbs)


class Nbe(_Nbm):
    """Class to handle NBM NBE report data"""

    report_type = "nbe"
    _parser = staticmethod(parse_nbe)



"""
Functions for parsing PIREPs
"""

# pylint: disable=too-many-boolean-expressions

# stdlib
from datetime import date
from typing import List, Optional, Tuple, Union

# module
from avwx.current.base import Reports, get_wx_codes
from avwx.parsing import core, sanitization
from avwx.static.core import CARDINALS, CLOUD_LIST, NA_UNITS
from avwx.structs import (
    Aircraft,
    Cloud,
    Code,
    Icing,
    Location,
    Number,
    PirepData,
    Timestamp,
    Turbulence,
    Units,
)

_UNITS = Units(**NA_UNITS)


def _root(item: str) -> Tuple[Optional[str], Optional[str]]:
    """Parses report root data including station and report type"""
    # pylint: disable=redefined-argument-from-local
    report_type = None
    station = None
    for item in item.split():
        if item in ("UA", "UUA"):
            report_type = item
        elif not station:
            station = item
    return station, report_type


def _location(item: str) -> Optional[Location]:
    """Convert a location element to a Location object"""
    items = item.split()
    if not items:
        return None
    station, direction, distance = None, None, None
    direction_number, distance_number = None, None
    if len(items) == 1:
        ilen = len(item)
        # MLB
        if ilen < 5:
            station = item
        # MKK360002 or KLGA220015
        elif ilen in (9, 10) and item[-6:].isdigit():
            station, direction, distance = item[:-6], item[-6:-3], item[-3:]
    # 10 WGON
    # 10 EAST
    # 15 SW LRP
    elif items[0].isdigit():
        if items[1] in CARDINALS:
            distance, direction = items[0], items[1]
            if len(items) == 3:
                station = items[2]
        else:
            station, direction, distance = items[1][-3:], items[1][:-3], items[0]
    # GON 270010
    elif items[1].isdigit():
        station, direction, distance = items[0], items[1][:3], items[1][3:]
    # Convert non-null elements
    if direction:
        direction_number = core.make_number(direction, literal=True)
    if distance:
        distance_number = core.make_number(distance)
    return Location(item, station, direction_number, distance_number)


def _time(item: str, target: date = None) -> Optional[Timestamp]:
    """Convert a time element to a Timestamp"""
    return core.make_timestamp(item, time_only=True, target_date=target)


def _altitude(item: str) -> Union[Optional[Number], str]:
    """Convert reporting altitude to a Number or string"""
    if item.isdigit():
        return core.make_number(item)
    return item


def _aircraft(item: str) -> Union[Aircraft, str]:
    """Returns the Aircraft from the ICAO code"""
    try:
        return Aircraft.from_icao(item)
    except ValueError:
        return item


def _clouds(item: str) -> List[Cloud]:
    """Convert cloud element to a list of Clouds"""
    clouds = item.replace(",", "").split()
    # BASES 004 TOPS 016
    # BASES SCT030 TOPS SCT058
    if "BASES" in clouds and "TOPS" in clouds:
        cloud_type = None
        base = clouds[clouds.index("BASES") + 1]
        top = clouds[clouds.index("TOPS") + 1]
        if not base.isdigit():
            cloud_type, base = base[:3], base[3:]
        if not top.isdigit():
            cloud_type, top = top[:3], top[3:]
        return [Cloud(item, cloud_type, base=int(base), top=int(top))]
    return [core.make_cloud(cloud) for cloud in clouds]


def _number(item: str) -> Optional[Number]:
    """Convert an element to a Number"""
    return core.make_number(item.strip("CF"), item)


def _find_floor_ceiling(
    items: List[str],
) -> Tuple[List[str], Optional[Number], Optional[Number]]:
    """Extracts the floor and ceiling from item list"""
    floor: Optional[Number] = None
    ceiling: Optional[Number] = None

    for i, item in enumerate(items):
        hloc = item.find("-")
        # TRACE RIME 070-090
        if hloc > -1 and item[:hloc].isdigit() and item[hloc + 1 :].isdigit():
            floor_str, ceiling_str = items.pop(i).split("-")
            floor = core.make_number(floor_str)
            ceiling = core.make_number(ceiling_str)
            break
        # CONT LGT CHOP BLO 250
        if item == "BLO":
            ceiling = core.make_number(items[i + 1])
            items = items[:i]
            break
        # LGT RIME 025
        if item.isdigit():
            num = core.make_number(item)
            floor, ceiling = num, num
            break
    return items, floor, ceiling


def _turbulence(item: str) -> Turbulence:
    """Convert reported turbulence to a Turbulence object"""
    items, floor, ceiling = _find_floor_ceiling(item.split())
    return Turbulence(
        severity=" ".join(items),
        floor=floor,
        ceiling=ceiling,
    )


def _icing(item: str) -> Icing:
    """Convert reported icing to an Icing object"""
    items, floor, ceiling = _find_floor_ceiling(item.split())
    severity = items.pop(0)
    return Icing(
        severity=severity,
        floor=floor,
        ceiling=ceiling,
        type=items[0] if items else None,
    )


def _remarks(item: str) -> str:
    """Returns the remarks. Reserved for later parsing"""
    return item


def _wx(item: str) -> Tuple[List[Code], Optional[Number], List[str]]:
    """Parses remaining weather elements"""
    # pylint: disable=redefined-argument-from-local
    other: List[str] = []
    flight_visibility = None
    for item in item.split():
        if len(item) > 2 and item.startswith("FV"):
            _, flight_visibility = core.get_visibility([item[2:]], _UNITS)
        else:
            other.append(item)
    other, wx_codes = get_wx_codes(other)
    return wx_codes, flight_visibility, other


def _sanitize_report_list(data: List[str]) -> List[str]:
    """Fixes report elements based on neighbor values"""
    for i, item in reversed(list(enumerate(data))):
        # Fix spaced cloud top Ex: BKN030 TOP045   BASE020 TOPS074
        # But not BASES SCT014 TOPS SCT021
        if (
            item.startswith("TOP")
            and item != "TOPS"
            and i > 0
            and len(data[i - 1]) >= 6
            and (data[i - 1][:3] in CLOUD_LIST or data[i - 1].startswith("BASE"))
        ):
            data[i - 1] += "-" + data.pop(i)
    wxdata = core.dedupe(data, only_neighbors=True)
    return wxdata


def sanitize(report: str) -> str:
    """Returns a sanitized report ready for parsing"""
    clean = sanitization.sanitize_report_string(report)
    data = _sanitize_report_list(clean.split())
    return " ".join(data)


def parse(report: str, issued: date = None) -> Optional[PirepData]:
    """Returns a PirepData object based on the given report"""
    # pylint: disable=too-many-locals,too-many-branches
    if not report:
        return None
    sanitized = sanitize(report)
    data = sanitized.split("/")
    station, report_type = _root(data.pop(0).strip())
    time, location, altitude, aircraft = None, None, None, None
    clouds, temperature, turbulence, other = None, None, None, None
    icing, remarks, flight_visibility, wx_codes = None, None, None, None
    for item in data:
        if not item or len(item) < 2:
            continue
        tag = item[:2]
        item = item[2:].strip()
        if tag == "TM":
            time = _time(item, issued)
        elif tag == "OV":
            location = _location(item)
        elif tag == "FL":
            altitude = _altitude(item)
        elif tag == "TP":
            aircraft = _aircraft(item)
        elif tag == "SK":
            clouds = _clouds(item)
        elif tag == "TA":
            temperature = _number(item)
        elif tag == "TB":
            turbulence = _turbulence(item)
        elif tag == "IC":
            icing = _icing(item)
        elif tag == "RM":
            remarks = _remarks(item)
        elif tag == "WX":
            wx_codes, flight_visibility, other = _wx(item)
    return PirepData(
        aircraft=aircraft,
        altitude=altitude,
        clouds=clouds,
        flight_visibility=flight_visibility,
        icing=icing,
        location=location,
        other=other or [],
        raw=report,
        remarks=remarks,
        sanitized=sanitized,
        station=station,
        temperature=temperature,
        time=time,
        turbulence=turbulence,
        type=report_type,
        wx_codes=wx_codes or [],
    )


class Pireps(Reports):
    """Class to handle pilot report data"""

    data: Optional[List[PirepData]] = None  # type: ignore

    @staticmethod
    def _report_filter(reports: List[str]) -> List[str]:
        """Removes AIREPs before updating raw_reports"""
        return [r for r in reports if not r.startswith("ARP")]

    async def _post_update(self):
        self.data = []
        for report in self.raw:
            self.data.append(parse(report, issued=self.issued))

    def _post_parse(self):
        self.data = []
        for report in self.raw:
            self.data.append(parse(report, issued=self.issued))

    @staticmethod
    def sanitize(report: str) -> str:
        """Sanitizes a PIREP string"""
        return sanitize(report)



"""
Contains functions for handling and translating remarks
"""

# pylint: disable=redefined-builtin

# stdlib
from contextlib import suppress
from typing import List, Optional, Tuple

# module
from avwx.parsing import core
from avwx.static.core import REMARKS_ELEMENTS, REMARKS_GROUPS, WX_TRANSLATIONS
from avwx.static.taf import PRESSURE_TENDENCIES
from avwx.structs import Code, FiveDigitCodes, Number, PressureTendency, RemarksData


Codes = List[str]


def decimal_code(code: str, repr: Optional[str] = None) -> Optional[Number]:
    """Parses a 4-digit decimal temperature representation

    Ex: 1045 -> -4.5    0237 -> 23.7
    """
    if not code:
        return None
    number = f"{'-' if code[0] == '1' else ''}{int(code[1:3])}.{code[3]}"
    return core.make_number(number, repr or code)


def temp_dew_decimal(codes: Codes) -> Tuple[Codes, Optional[Number], Optional[Number]]:
    """Returns the decimal temperature and dewpoint values"""
    temp, dew = None, None
    for i, code in reversed(list(enumerate(codes))):
        if len(code) in [5, 9] and code[0] == "T" and code[1:].isdigit():
            codes.pop(i)
            temp, dew = decimal_code(code[1:5]), decimal_code(code[5:])
            break
    return codes, temp, dew


def temp_minmax(codes: Codes) -> Tuple[Codes, Optional[Number], Optional[Number]]:
    """Returns the 24-hour minimum and maximum temperatures"""
    maximum, minimum = None, None
    for i, code in enumerate(codes):
        if len(code) == 9 and code[0] == "4" and code.isdigit():
            maximum, minimum = decimal_code(code[1:5]), decimal_code(code[5:])
            codes.pop(i)
            break
    return codes, maximum, minimum


def precip_snow(codes: Codes) -> Tuple[Codes, Optional[Number], Optional[Number]]:
    """Returns the hourly precipitation and snow depth"""
    precip, snow = None, None
    for i, code in reversed(list(enumerate(codes))):
        if len(code) != 5:
            continue
        # P0213
        if code[0] == "P" and code[1:].isdigit():
            precip = core.make_number(f"{code[1:3]}.{code[3:]}", code)
            codes.pop(i)
        # 4/012
        elif code[:2] == "4/" and code[2:].isdigit():
            snow = core.make_number(code[2:], code)
            codes.pop(i)
    return codes, precip, snow


def sea_level_pressure(codes: Codes) -> Tuple[Codes, Optional[Number]]:
    """Returns the sea level pressure always in hPa"""
    sea = None
    for i, code in enumerate(codes):
        if len(code) == 6 and code.startswith("SLP") and code[-3:].isdigit():
            value = f"{'9' if int(code[-3]) > 4 else '10'}{code[-3:-1]}.{code[-1]}"
            sea = core.make_number(value, code)
            codes.pop(i)
            break
    return codes, sea


def parse_pressure(code: str) -> PressureTendency:
    """Parse a 5-digit pressure tendency"""
    return PressureTendency(
        repr=code,
        tendency=PRESSURE_TENDENCIES[code[1]],
        change=float(f"{code[2:4]}.{code[4]}"),
    )


def parse_precipitation(code: str) -> Optional[Number]:
    """Parse a 5-digit precipitation amount"""
    return core.make_number(f"{code[1:3]}.{code[3:]}", code)


def five_digit_codes(codes: Codes) -> Tuple[Codes, FiveDigitCodes]:
    """Returns  a 5-digit min/max temperature code"""
    values = FiveDigitCodes()
    for i, code in reversed(list(enumerate(codes))):
        if len(code) == 5 and code.isdigit():
            key = int(code[0])
            if key == 1:
                values.maximum_temperature_6 = decimal_code(code[1:], code)
            elif key == 2:
                values.minimum_temperature_6 = decimal_code(code[1:], code)
            elif key == 5:
                values.pressure_tendency = parse_pressure(code)
            elif key == 6:
                values.precip_36_hours = parse_precipitation(code)
            elif key == 7:
                values.precip_24_hours = parse_precipitation(code)
            elif key == 9:
                values.sunshine_minutes = core.make_number(code[2:], code)
            else:
                continue
            codes.pop(i)
    return codes, values


def find_codes(rmk: str) -> Tuple[Codes, List[Code]]:
    """Find a remove known static codes from the starting remarks list"""
    ret = []
    for key, value in REMARKS_GROUPS.items():
        if key in rmk:
            ret.append(Code(key, value))
            rmk.replace(key, "")
    codes = [i for i in rmk.split() if i]
    for i, code in reversed(list(enumerate(codes))):
        with suppress(KeyError):
            ret.append(Code(code, REMARKS_ELEMENTS[code]))
            codes.pop(i)
        # Weather began/ended
        if (
            len(code) == 5
            and code[2] in ("B", "E")
            and code[3:].isdigit()
            and code[:2] in WX_TRANSLATIONS
        ):
            state = "began" if code[2] == "B" else "ended"
            value = f"{WX_TRANSLATIONS[code[:2]]} {state} at :{code[3:]}"
            ret.append(Code(code, value))
            codes.pop(i)
    ret.sort(key=lambda x: x.repr)
    return codes, ret


def parse(rmk: str) -> Optional[RemarksData]:
    """Finds temperature and dewpoint decimal values from the remarks"""
    if not rmk:
        return None
    codes, parsed_codes = find_codes(rmk)
    codes, temperature, dewpoint = temp_dew_decimal(codes)
    codes, max_temp, min_temp = temp_minmax(codes)
    codes, precip, snow = precip_snow(codes)
    codes, sea = sea_level_pressure(codes)
    codes, fivedigits = five_digit_codes(codes)
    return RemarksData(
        codes=parsed_codes,
        dewpoint_decimal=dewpoint,
        temperature_decimal=temperature,
        minimum_temperature_6=fivedigits.minimum_temperature_6,
        minimum_temperature_24=min_temp,
        maximum_temperature_6=fivedigits.maximum_temperature_6,
        maximum_temperature_24=max_temp,
        pressure_tendency=fivedigits.pressure_tendency,
        precip_36_hours=fivedigits.precip_36_hours,
        precip_24_hours=fivedigits.precip_24_hours,
        sunshine_minutes=fivedigits.sunshine_minutes,
        precip_hourly=precip,
        snow_depth=snow,
        sea_level_pressure=sea,
    )



"""
Contains the core parsing and indent functions of avwx
"""

# pylint: disable=redefined-builtin

# stdlib
import re
import datetime as dt
import math
from calendar import monthrange
from contextlib import suppress
from copy import copy
from typing import Any, Iterable, List, Optional, Tuple, Union

# library
from dateutil.relativedelta import relativedelta

# module
from avwx.static.core import (
    CARDINALS,
    CLOUD_LIST,
    FRACTIONS,
    NUMBER_REPL,
    SPECIAL_NUMBERS,
)
from avwx.structs import Cloud, Fraction, Number, Timestamp, Units


def dedupe(items: Iterable[Any], only_neighbors: bool = False) -> List[Any]:
    """Deduplicate a list while keeping order

    If only_neighbors is True, dedupe will only check neighboring values
    """
    ret: List[Any] = []
    for item in items:
        if (only_neighbors and ret and ret[-1] != item) or item not in ret:
            ret.append(item)
    return ret


def is_unknown(value: str) -> bool:
    """Returns True if val represents and unknown value"""
    if not isinstance(value, str):
        raise TypeError
    if not value or value.upper() in ("UNKN",):
        return True
    for char in value:
        if char not in ("/", "X", "."):
            break
    else:
        return True
    return False


def get_digit_list(data: List[str], from_index: int) -> Tuple[List[str], List[str]]:
    """Returns a list of items removed from a given list of strings
    that are all digits from 'from_index' until hitting a non-digit item
    """
    ret = []
    data.pop(from_index)
    while len(data) > from_index and data[from_index].isdigit():
        ret.append(data.pop(from_index))
    return data, ret


def unpack_fraction(num: str) -> str:
    """Returns unpacked fraction string 5/2 -> 2 1/2"""
    numbers = [int(n) for n in num.split("/") if n]
    if not (len(numbers) == 2 and numbers[0] > numbers[1]):
        return num
    numerator, denominator = numbers
    over = numerator // denominator
    rem = numerator % denominator
    return f"{over} {rem}/{denominator}"


def remove_leading_zeros(num: str) -> str:
    """Strips zeros while handling -, M, and empty strings"""
    if not num:
        return num
    if num.startswith("M"):
        ret = "M" + num[1:].lstrip("0")
    elif num.startswith("-"):
        ret = "-" + num[1:].lstrip("0")
    else:
        ret = num.lstrip("0")
    return "0" if ret in ("", "M", "-") else ret


SPOKEN_POSTFIX = (
    (" zero zero zero", " thousand"),
    (" zero zero", " hundred"),
)


def spoken_number(num: str, literal: bool = False) -> str:
    """Returns the spoken version of a number

    If literal, no conversion to hundreds/thousands

    Ex: 1.2 -> one point two
        1 1/2 -> one and one half
        25000 -> two five thousand
    """
    ret = []
    for part in num.split():
        if part in FRACTIONS:
            ret.append(FRACTIONS[part])
        else:
            val = " ".join(NUMBER_REPL[char] for char in part if char in NUMBER_REPL)
            if not literal:
                for target, replacement in SPOKEN_POSTFIX:
                    if val.endswith(target):
                        val = val[: -len(target)] + replacement
            ret.append(val)
    return " and ".join(ret)


def make_fraction(num: str, repr: str = None, literal: bool = False) -> Fraction:
    """Returns a fraction dataclass for numbers with / in them"""
    num_str, den_str = num.split("/")
    # 2-1/2 but not -2 1/2
    if "-" in num_str and not num_str.startswith("-"):
        num_str = num_str.replace("-", " ")
    denominator = int(den_str)
    # Multiply multi-digit numerator
    if len(num_str) > 1:
        numerator = int(num_str[:-1]) * denominator + int(num_str[-1])
        num = f"{numerator}/{denominator}"
    else:
        numerator = int(num_str)
    value = numerator / denominator
    unpacked = unpack_fraction(num)
    spoken = spoken_number(unpacked, literal)
    return Fraction(repr or num, value, spoken, numerator, denominator, unpacked)


def make_number(
    num: Optional[str],
    repr: str = None,
    speak: str = None,
    literal: bool = False,
    special: dict = None,
) -> Optional[Number]:
    """Returns a Number or Fraction dataclass for a number string

    If literal, spoken string will not convert to hundreds/thousands

    NOTE: Numerators are assumed to have a single digit. Additional are whole numbers
    """
    if not num or is_unknown(num):
        return None
    # Check special
    with suppress(KeyError):
        value, spoken = (special or {}).get(num) or SPECIAL_NUMBERS[num]
        return Number(repr or num, value, spoken)
    # Check cardinal direction
    if num in CARDINALS:
        if not repr:
            repr = num
        num = str(CARDINALS[num])
    # Remove spurious characters from the end
    num = num.rstrip("M.")
    num = num.replace("O", "0")
    num = num.replace("+", "")
    # Create Fraction
    if "/" in num:
        return make_fraction(num, repr, literal)
    # Handle Minus values with errors like 0M04
    if "M" in num:
        val_str = num.replace("M", "-")
        while val_str[0] != "-":
            val_str = val_str[1:]
    else:
        val_str = num
    # Create Number
    if not val_str:
        return None
    if "." in num:
        value = float(val_str)
        # Overwrite float 0 due to "0.0" literal
        if not value:
            value = 0
    else:
        value = int(val_str)
    return Number(repr or num, value, spoken_number(speak or str(value), literal))


def find_first_in_list(txt: str, str_list: List[str]) -> int:
    """Returns the index of the earliest occurrence of an item from a list in a string

    Ex: find_first_in_list('foobar', ['bar', 'fin']) -> 3
    """
    start = len(txt) + 1
    for item in str_list:
        if start > txt.find(item) > -1:
            start = txt.find(item)
    return start if len(txt) + 1 > start > -1 else -1


def is_timestamp(item: str) -> bool:
    """Returns True if the item matches the timestamp format"""
    return len(item) == 7 and item[-1] == "Z" and item[:-1].isdigit()


def is_timerange(item: str) -> bool:
    """Returns True if the item is a TAF to-from time range"""
    return (
        len(item) == 9 and item[4] == "/" and item[:4].isdigit() and item[5:].isdigit()
    )


def is_possible_temp(temp: str) -> bool:
    """Returns True if all characters are digits or 'M' (for minus)"""
    for char in temp:
        if not (char.isdigit() or char == "M"):
            return False
    return True


_numeric = Union[int, float]


def relative_humidity(
    temperature: _numeric, dewpoint: _numeric, unit: str = "C"
) -> float:
    """Calculates the relative humidity as a 0 to 1 percentage"""

    def saturation(value: Union[int, float]) -> float:
        """Returns the saturation vapor pressure without the C constant for humidity calc"""
        return math.exp((17.67 * value) / (243.5 + value))

    if unit == "F":
        dewpoint = (dewpoint - 32) * 5 / 9
        temperature = (temperature - 32) * 5 / 9
    return saturation(dewpoint) / saturation(temperature)


# https://aviation.stackexchange.com/questions/47971/how-do-i-calculate-density-altitude-by-hand


def pressure_altitude(pressure: float, altitude: _numeric, unit: str = "inHg") -> int:
    """Calculates the pressure altitude in feet. Converts pressure units"""
    if unit == "hPa":
        pressure *= 0.02953
    return round((29.92 - pressure) * 1000 + altitude)


def density_altitude(
    pressure: float, temperature: _numeric, altitude: _numeric, units: Units
) -> int:
    """Calculates the density altitude in feet. Converts pressure and temperature units"""
    if units.temperature == "F":
        temperature = (temperature - 32) * 5 / 9
    if units.altimeter == "hPa":
        pressure *= 0.02953
    pressure_alt = pressure_altitude(pressure, altitude)
    standard = 15 - (2 * altitude / 1000)
    return round(((temperature - standard) * 120) + pressure_alt)


def get_station_and_time(
    data: List[str],
) -> Tuple[List[str], Optional[str], Optional[str]]:
    """Returns the report list and removed station ident and time strings"""
    if not data:
        return data, None, None
    station = data.pop(0)
    if not data:
        return data, station, None
    q_time, r_time = data[0], None
    if data and q_time.endswith("Z") and q_time[:-1].isdigit():
        r_time = data.pop(0)
    elif data and len(q_time) == 6 and q_time.isdigit():
        r_time = data.pop(0) + "Z"
    return data, station, r_time


def is_wind(text: str) -> bool:
    """Returns True if the text is likely a normal wind element"""
    # Ignore wind shear
    if text.startswith("WS"):
        return False
    # 09010KT, 09010G15KT
    if len(text) > 4:
        for ending in ("KT", "KTS", "MPS", "KMH"):
            if text.endswith(ending):
                return True
    # 09010  09010G15 VRB10
    if not (
        len(text) == 5
        or (len(text) >= 8 and text.find("G") != -1)
        and text.find("/") == -1
    ):
        return False
    return text[:5].isdigit() or (text.startswith("VRB") and text[3:5].isdigit())


VARIABLE_DIRECTION_PATTERN = re.compile(r"\d{3}V\d{3}")


def is_variable_wind_direction(text: str) -> bool:
    """Returns True if element looks like 350V040"""
    if len(text) < 7:
        return False
    return VARIABLE_DIRECTION_PATTERN.match(text[:7]) is not None


def separate_wind(text: str) -> Tuple[str, str, str]:
    """Extracts the direction, speed, and gust from a wind element"""
    direction, speed, gust = "", "", ""
    # Remove gust
    if "G" in text:
        g_index = text.find("G")
        start, end = g_index + 1, g_index + 3
        # 16006GP99KT ie gust greater than
        if "GP" in text:
            end += 1
        gust = text[start:end]
        text = text[:g_index] + text[end:]
    if text:
        # 10G18KT
        if len(text) == 2:
            speed = text
        else:
            direction = text[:3]
            speed = text[3:]
    return direction, speed, gust


def get_wind(
    data: List[str], units: Units
) -> Tuple[
    List[str],
    Optional[Number],
    Optional[Number],
    Optional[Number],
    List[Number],
]:
    """Returns the report list and removed:

    Direction string, speed string, gust string, variable direction list
    """
    direction, speed, gust = "", "", ""
    variable: List[Number] = []
    if data:
        item = copy(data[0])
        if is_wind(item):
            # Select and remove unit in order of frequency
            if item.endswith("KT"):
                item = item.replace("KT", "")
            elif item.endswith("KTS"):
                item = item.replace("KTS", "")
            elif item.endswith("MPS"):
                units.wind_speed = "m/s"
                item = item.replace("MPS", "")
            elif item.endswith("KMH"):
                units.wind_speed = "km/h"
                item = item.replace("KMH", "")
            direction, speed, gust = separate_wind(item)
            data.pop(0)
    # Separated Gust
    if data and 1 < len(data[0]) < 4 and data[0][0] == "G" and data[0][1:].isdigit():
        gust = data.pop(0)[1:]
    # Variable Wind Direction
    if data and is_variable_wind_direction(data[0]):
        for item in data.pop(0).split("V"):
            value = make_number(item, speak=item, literal=True)
            if value is not None:
                variable.append(value)
    # Convert to Number
    direction_value = make_number(direction, speak=direction, literal=True)
    speed_value = make_number(speed.strip("BV"))
    gust_value = make_number(gust)
    return data, direction_value, speed_value, gust_value, variable


def get_visibility(data: List[str], units: Units) -> Tuple[List[str], Optional[Number]]:
    """Returns the report list and removed visibility string"""
    visibility = ""
    if data:
        item = copy(data[0])
        # Vis reported in statue miles
        if item.endswith("SM"):  # 10SM
            if item in ("P6SM", "M1/4SM", "M1/8SM"):
                visibility = item[:-2]
            elif item[:-2].isdigit():
                visibility = str(int(item[:-2]))
            elif "/" in item:
                visibility = item[: item.find("SM")]  # 1/2SM
            data.pop(0)
            units.visibility = "sm"
        # Vis reported in meters
        elif len(item) == 4 and item.isdigit():
            visibility = data.pop(0)
            units.visibility = "m"
        elif (
            7 >= len(item) >= 5
            and item[:4].isdigit()
            and (item[4] in ["M", "N", "S", "E", "W"] or item[4:] == "NDV")
        ):
            visibility = data.pop(0)[:4]
            units.visibility = "m"
        elif len(item) == 5 and item[1:].isdigit() and item[0] in ["M", "P", "B"]:
            visibility = data.pop(0)[1:]
            units.visibility = "m"
        elif item.endswith("KM") and item[:-2].isdigit():
            visibility = item[:-2] + "000"
            data.pop(0)
            units.visibility = "m"
        # Vis statute miles but split Ex: 2 1/2SM
        elif (
            len(data) > 1
            and data[1].endswith("SM")
            and "/" in data[1]
            and item.isdigit()
        ):
            vis1 = data.pop(0)  # 2
            vis2 = data.pop(0).replace("SM", "")  # 1/2
            visibility = str(int(vis1) * int(vis2[2]) + int(vis2[0])) + vis2[1:]  # 5/2
            units.visibility = "sm"
    return data, make_number(visibility)


def sanitize_cloud(cloud: str) -> str:
    """Fix rare cloud layer issues"""
    if len(cloud) < 4:
        return cloud
    if not cloud[3].isdigit() and cloud[3] not in ("/", "-"):
        # Bad "O": FEWO03 -> FEW003
        if cloud[3] == "O":
            cloud = cloud[:3] + "0" + cloud[4:]
        # Move modifiers to end: BKNC015 -> BKN015C
        elif cloud[3] != "U" and not cloud.startswith("BASE"):
            cloud = cloud[:3] + cloud[4:] + cloud[3]
    return cloud


def _null_or_int(val: Optional[str]) -> Optional[int]:
    """Nullify unknown elements and convert ints"""
    if not isinstance(val, str) or is_unknown(val):
        return None
    return int(val)


_TOP_OFFSETS = ("-TOPS", "-TOP")


def make_cloud(cloud: str) -> Cloud:
    """Returns a Cloud dataclass for a cloud string

    This function assumes the input is potentially valid
    """
    raw_cloud = cloud
    type = ""
    base: Optional[str] = None
    top: Optional[str] = None
    modifier: Optional[str] = None
    cloud = sanitize_cloud(cloud).replace("/", "")
    # Separate top
    for target in _TOP_OFFSETS:
        topi = cloud.find(target)
        if topi > -1:
            top, cloud = cloud[topi + len(target) :], cloud[:topi]
            break
    # Separate type
    ## BASE027
    if cloud.startswith("BASES"):
        cloud = cloud[5:]
    elif cloud.startswith("BASE"):
        cloud = cloud[4:]
    ## VV003
    elif cloud.startswith("VV"):
        type, cloud = cloud[:2], cloud[2:]
    ## FEW010
    elif len(cloud) >= 3 and cloud[:3] in CLOUD_LIST:
        type, cloud = cloud[:3], cloud[3:]
    ## BKN-OVC065
    if len(cloud) > 4 and cloud[0] == "-" and cloud[1:4] in CLOUD_LIST:
        type += cloud[:4]
        cloud = cloud[4:]
    # Separate base
    if len(cloud) >= 3 and cloud[:3].isdigit():
        base, cloud = cloud[:3], cloud[3:]
    elif len(cloud) >= 4 and cloud[:4] == "UNKN":
        cloud = cloud[4:]
    # Remainder is considered modifiers
    if cloud:
        modifier = cloud
    # Make Cloud
    return Cloud(
        raw_cloud, type or None, _null_or_int(base), _null_or_int(top), modifier
    )


def get_clouds(data: List[str]) -> Tuple[List[str], list]:
    """Returns the report list and removed list of split cloud layers"""
    clouds = []
    for i, item in reversed(list(enumerate(data))):
        if item[:3] in CLOUD_LIST or item[:2] == "VV":
            cloud = data.pop(i)
            clouds.append(make_cloud(cloud))
    # Attempt cloud sort. Fails if None values are present
    try:
        clouds.sort(key=lambda cloud: (cloud.base, cloud.type))
    except TypeError:
        clouds.reverse()  # Restores original report order
    return data, clouds


def get_flight_rules(visibility: Optional[Number], ceiling: Optional[Cloud]) -> int:
    """Returns int based on current flight rules from parsed METAR data

    0=VFR, 1=MVFR, 2=IFR, 3=LIFR

    Note: Common practice is to report no higher than IFR if visibility unavailable
    """
    # Parse visibility
    vis: Union[int, float]
    if visibility is None:
        vis = 2
    elif visibility.repr == "CAVOK" or visibility.repr.startswith("P6"):
        vis = 10
    elif visibility.repr.startswith("M"):
        vis = 0
    elif visibility.value is None:
        vis = 2
    # Convert meters to miles
    elif len(visibility.repr) == 4:
        vis = (visibility.value or 0) * 0.000621371
    else:
        vis = visibility.value or 0
    # Parse ceiling
    cld = (ceiling.base if ceiling else 99) or 99
    # Determine flight rules
    if (vis <= 5) or (cld <= 30):
        if (vis < 3) or (cld < 10):
            if (vis < 1) or (cld < 5):
                return 3  # LIFR
            return 2  # IFR
        return 1  # MVFR
    return 0  # VFR


def get_ceiling(clouds: List[Cloud]) -> Optional[Cloud]:
    """Returns ceiling layer from Cloud-List or None if none found

    Assumes that the clouds are already sorted lowest to highest

    Only 'Broken', 'Overcast', and 'Vertical Visibility' are considered ceilings

    Prevents errors due to lack of cloud information (eg. '' or 'FEW///')
    """
    for cloud in clouds:
        if cloud.base and cloud.type in ("OVC", "BKN", "VV"):
            return cloud
    return None


def parse_date(
    date: str,
    hour_threshold: int = 200,
    time_only: bool = False,
    target: dt.date = None,
) -> Optional[dt.datetime]:
    """Parses a report timestamp in ddhhZ or ddhhmmZ format

    If time_only, assumes hhmm format with current or previous day

    This function assumes the given timestamp is within the hour threshold from current date
    """
    # pylint: disable=too-many-branches
    # Format date string
    date = date.strip("Z")
    if not date.isdigit():
        return None
    if time_only:
        if len(date) != 4:
            return None
        index_hour = 0
    else:
        if len(date) == 4:
            date += "00"
        if len(date) != 6:
            return None
        index_hour = 2
    # Create initial guess
    if target:
        target = dt.datetime(
            target.year, target.month, target.day, tzinfo=dt.timezone.utc
        )
    else:
        target = dt.datetime.now(tz=dt.timezone.utc)
    day = target.day if time_only else int(date[0:2])
    hour = int(date[index_hour : index_hour + 2])
    # Handle situation where next month has less days than current month
    # Shifted value makes sure that a month shift doesn't happen twice
    shifted = False
    if day > monthrange(target.year, target.month)[1]:
        target += relativedelta(months=-1)
        shifted = True
    try:
        guess = target.replace(
            day=day,
            hour=hour % 24,
            minute=int(date[index_hour + 2 : index_hour + 4]) % 60,
            second=0,
            microsecond=0,
        )
    except ValueError:
        return None
    # Handle overflow hour
    if hour > 23:
        guess += dt.timedelta(days=1)
    # Handle changing months if not already shifted
    if not shifted:
        hourdiff = (guess - target) / dt.timedelta(minutes=1) / 60
        if hourdiff > hour_threshold:
            guess += relativedelta(months=-1)
        elif hourdiff < -hour_threshold:
            guess += relativedelta(months=+1)
    return guess


def make_timestamp(
    timestamp: Optional[str], time_only: bool = False, target_date: dt.date = None
) -> Optional[Timestamp]:
    """Returns a Timestamp dataclass for a report timestamp in ddhhZ or ddhhmmZ format"""
    if not timestamp:
        return None
    date_obj = parse_date(timestamp, time_only=time_only, target=target_date)
    return Timestamp(timestamp, date_obj)



"""
Contains METAR-specific functions for report parsing
"""

# pylint: disable=invalid-overridden-method

# stdlib
from contextlib import suppress
from datetime import date, datetime, timedelta, timezone
from typing import List, Tuple, Optional

# module
from avwx.current.base import Report, get_wx_codes
from avwx.parsing import core, remarks, sanitization, speech, summary
from avwx.parsing.translate.metar import translate_metar
from avwx.service import NOAA
from avwx.static.core import FLIGHT_RULES, IN_UNITS, NA_UNITS
from avwx.static.metar import METAR_RMK
from avwx.station import uses_na_format, valid_station
from avwx.structs import (
    Code,
    MetarData,
    MetarTrans,
    Number,
    RemarksData,
    RunwayVisibility,
    Units,
)


def get_remarks(txt: str) -> Tuple[List[str], str]:
    """Returns the report split into components and the remarks string

    Remarks can include items like RMK and on, NOSIG and on, and BECMG and on
    """
    txt = txt.replace("?", "").strip()
    # First look for Altimeter in txt
    alt_index = len(txt) + 1
    for item in [" A2", " A3", " Q1", " Q0", " Q9"]:
        index = txt.find(item)
        if len(txt) - 6 > index > -1 and txt[index + 2 : index + 6].isdigit():
            alt_index = index
    # Then look for earliest remarks 'signifier'
    sig_index = core.find_first_in_list(txt, METAR_RMK)
    if sig_index == -1:
        sig_index = len(txt) + 1
    if sig_index > alt_index > -1:
        return txt[: alt_index + 6].strip().split(), txt[alt_index + 7 :]
    if alt_index > sig_index > -1:
        return txt[:sig_index].strip().split(), txt[sig_index + 1 :]
    return txt.strip().split(), ""


def is_runway_visibility(item: str) -> bool:
    """Returns True if the item is a runway visibility range string"""
    return (
        len(item) > 4
        and item[0] == "R"
        and (item[3] == "/" or item[4] == "/")
        and item[1:3].isdigit()
    )


_RVR_CODES = {
    "M": "less than",
    "P": "greater than",
    "U": "increasing",
    "D": "decreasing",
    "N": "no change",
    "V": "variable",
}


def _parse_rvr_number(value: str) -> Optional[Number]:
    raw, prefix = value, None
    with suppress(KeyError):
        prefix = _RVR_CODES[value[0]]
        value = value[1:]
    number = core.make_number(value, raw)
    if number is not None and prefix is not None:
        number.spoken = prefix + " " + number.spoken
        number.value = None
    return number


def parse_runway_visibility(value: str) -> RunwayVisibility:
    """Parse a runway visibility range string"""
    raw, trend = value, None
    value = value.replace("FT", "")
    with suppress(KeyError):
        trend = Code(value[-1], _RVR_CODES[value[-1]])
        value = value[:-1]
    runway, value, *_ = value[1:].split("/")
    if value:
        possible_numbers = [_parse_rvr_number(n) for n in value.split("V")]
        numbers = [n for n in possible_numbers if n is not None]
        visibility = None
        if len(numbers) == 1:
            visibility = numbers.pop()
    else:
        visibility, numbers = None, []
    return RunwayVisibility(
        repr=raw,
        runway=runway,
        visibility=visibility,
        variable_visibility=numbers,
        trend=trend,
    )


def get_runway_visibility(data: List[str]) -> Tuple[List[str], List[RunwayVisibility]]:
    """Returns the report list and the remove runway visibility list"""
    runway_vis = []
    for i, item in reversed(list(enumerate(data))):
        if is_runway_visibility(item):
            runway_vis.append(parse_runway_visibility(data.pop(i)))
    runway_vis.sort(key=lambda x: x.runway)
    return data, runway_vis


def parse_altimeter(value: str) -> Optional[Number]:
    """Parse an altimeter string into a Number"""
    if not value or len(value) < 4:
        return None
    # QNH3003INS
    if len(value) >= 7 and value.endswith("INS"):
        return core.make_number(value[-7:-5] + "." + value[-5:-3], value, literal=True)
    number = value.replace(".", "")
    # Q1000/10
    if "/" in number:
        number = number.split("/")[0]
    if number.startswith("QNH"):
        number = "Q" + number[1:]
    if not (len(number) in (4, 5) and number[-4:].isdigit()):
        return None
    number = number.lstrip("AQ")
    if number[0] in ("2", "3"):
        number = number[:2] + "." + number[2:]
    elif number[0] not in ("0", "1"):
        return None
    return core.make_number(number, value, number, literal=True)


def get_altimeter(
    data: List[str], units: Units, version: str = "NA"
) -> Tuple[List[str], Optional[Number]]:
    """Returns the report list and the removed altimeter item

    Version is 'NA' (North American / default) or 'IN' (International)
    """
    values: List[Number] = []
    for _ in range(2):
        if not data:
            break
        value = parse_altimeter(data[-1])
        if value is None:
            break
        values.append(value)
        data.pop(-1)
    if not values:
        return data, None
    values.sort(key=lambda x: x.value or 0)
    altimeter = values[0 if version == "NA" else -1]
    if altimeter.value is not None:
        units.altimeter = "inHg" if altimeter.value < 100 else "hPa"
    return data, altimeter


def get_temp_and_dew(
    data: List[str],
) -> Tuple[List[str], Optional[Number], Optional[Number]]:
    """Returns the report list and removed temperature and dewpoint strings"""
    for i, item in reversed(list(enumerate(data))):
        if "/" in item:
            # ///07
            if item[0] == "/":
                item = "/" + item.lstrip("/")
            # 07///
            elif item[-1] == "/":
                item = item.rstrip("/") + "/"
            tempdew = item.split("/")
            if len(tempdew) != 2:
                continue
            valid = True
            for j, temp in enumerate(tempdew):
                if temp in ["MM", "XX"]:
                    tempdew[j] = ""
                elif not core.is_possible_temp(temp):
                    valid = False
                    break
            if valid:
                data.pop(i)
                temp, dew = tempdew
                return data, core.make_number(temp), core.make_number(dew)
    return data, None, None


def get_relative_humidity(
    temperature: Optional[Number],
    dewpoint: Optional[Number],
    remarks_info: Optional[RemarksData],
    units: Units,
) -> Optional[float]:
    """Calculates relative humidity from preferred temperature and dewpoint"""
    if remarks_info is not None:
        temp = remarks_info.temperature_decimal or temperature
        dew = remarks_info.dewpoint_decimal or dewpoint
    else:
        temp = temperature
        dew = dewpoint
    if temp is None or temp.value is None:
        return None
    if dew is None or dew.value is None:
        return None
    return core.relative_humidity(temp.value, dew.value, units.temperature)


def sanitize(report: str) -> Tuple[str, str, List[str]]:
    """Returns a sanitized report, remarks, and elements ready for parsing"""
    clean = sanitization.sanitize_report_string(report)
    data, remark_str = get_remarks(clean)
    data = core.dedupe(data)
    data = sanitization.sanitize_report_list(data)
    clean = " ".join(data)
    if remark_str:
        clean += " " + remark_str
    return clean, remark_str, data


def parse(
    station: str, report: str, issued: date = None
) -> Tuple[Optional[MetarData], Optional[Units]]:
    """Returns MetarData and Units dataclasses with parsed data and their associated units"""
    valid_station(station)
    if not report:
        return None, None
    parser = parse_na if uses_na_format(station[:2]) else parse_in
    return parser(report, issued)


def parse_na(report: str, issued: date = None) -> Tuple[MetarData, Units]:
    """Parser for the North American METAR variant"""
    # pylint: disable=too-many-locals
    units = Units(**NA_UNITS)
    sanitized, remarks_str, data = sanitize(report)
    data, station, time = core.get_station_and_time(data)
    data, runway_visibility = get_runway_visibility(data)
    data, clouds = core.get_clouds(data)
    (
        data,
        wind_direction,
        wind_speed,
        wind_gust,
        wind_variable_direction,
    ) = core.get_wind(data, units)
    data, altimeter = get_altimeter(data, units, "NA")
    data, visibility = core.get_visibility(data, units)
    data, temperature, dewpoint = get_temp_and_dew(data)
    condition = core.get_flight_rules(visibility, core.get_ceiling(clouds))
    other, wx_codes = get_wx_codes(data)
    remarks_info = remarks.parse(remarks_str)
    humidity = get_relative_humidity(temperature, dewpoint, remarks_info, units)
    struct = MetarData(
        altimeter=altimeter,
        clouds=clouds,
        dewpoint=dewpoint,
        flight_rules=FLIGHT_RULES[condition],
        other=other,
        raw=report,
        relative_humidity=humidity,
        remarks_info=remarks_info,
        remarks=remarks_str,
        runway_visibility=runway_visibility,
        sanitized=sanitized,
        station=station,
        temperature=temperature,
        time=core.make_timestamp(time, target_date=issued),
        visibility=visibility,
        wind_direction=wind_direction,
        wind_gust=wind_gust,
        wind_speed=wind_speed,
        wind_variable_direction=wind_variable_direction,
        wx_codes=wx_codes,
    )
    return struct, units


def parse_in(report: str, issued: date = None) -> Tuple[MetarData, Units]:
    """Parser for the International METAR variant"""
    # pylint: disable=too-many-locals
    units = Units(**IN_UNITS)
    sanitized, remarks_str, data = sanitize(report)
    data, station, time = core.get_station_and_time(data)
    data, runway_visibility = get_runway_visibility(data)
    if "CAVOK" not in data:
        data, clouds = core.get_clouds(data)
    (
        data,
        wind_direction,
        wind_speed,
        wind_gust,
        wind_variable_direction,
    ) = core.get_wind(data, units)
    data, altimeter = get_altimeter(data, units, "IN")
    if "CAVOK" in data:
        visibility = core.make_number("CAVOK")
        clouds = []
        data.remove("CAVOK")
    else:
        data, visibility = core.get_visibility(data, units)
    data, temperature, dewpoint = get_temp_and_dew(data)
    condition = core.get_flight_rules(visibility, core.get_ceiling(clouds))
    other, wx_codes = get_wx_codes(data)
    remarks_info = remarks.parse(remarks_str)
    humidity = get_relative_humidity(temperature, dewpoint, remarks_info, units)
    struct = MetarData(
        altimeter=altimeter,
        clouds=clouds,
        dewpoint=dewpoint,
        flight_rules=FLIGHT_RULES[condition],
        other=other,
        raw=report,
        relative_humidity=humidity,
        remarks_info=remarks_info,
        remarks=remarks_str,
        runway_visibility=runway_visibility,
        sanitized=sanitized,
        station=station,
        temperature=temperature,
        time=core.make_timestamp(time, target_date=issued),
        visibility=visibility,
        wind_direction=wind_direction,
        wind_gust=wind_gust,
        wind_speed=wind_speed,
        wind_variable_direction=wind_variable_direction,
        wx_codes=wx_codes,
    )
    return struct, units


class Metar(Report):
    """Class to handle METAR report data"""

    data: Optional[MetarData] = None
    translations: Optional[MetarTrans] = None

    async def _pull_from_default(self) -> None:
        """Checks for a more recent report from NOAA. Only sync"""
        service = NOAA(self.__class__.__name__.lower())
        if self.icao is None:
            return
        report = await service.async_fetch(self.icao)
        if report is not None:
            data, units = parse(self.icao, report, self.issued)
            if not data or data.time is None or data.time.dt is None:
                return
            if (
                not self.data
                or self.data.time is None
                or self.data.time.dt is None
                or data.time.dt > self.data.time.dt
            ):
                self.data, self.units = data, units
                self.source = service.root

    @property
    def _should_check_default(self) -> bool:
        """Returns True if pulled from regional source and potentially out of date"""
        if isinstance(self.service, NOAA) or self.source is None:
            return False

        if self.data is None or self.data.time is None or self.data.time.dt is None:
            return True
        time_since = datetime.now(tz=timezone.utc) - self.data.time.dt
        return time_since > timedelta(minutes=90)

    def _calculate_altitudes(self):
        """Adds the pressure and density altitudes to data if all fields are available"""
        if self.data is None:
            return
        # Select decimal temperature if available
        temp = self.data.temperature
        if self.data.remarks_info is not None:
            temp = self.data.remarks_info.temperature_decimal or temp
        alt = self.data.altimeter
        if temp is None or temp.value is None or alt is None or alt.value is None:
            return
        alt, temp = alt.value, temp.value
        elev = self.station.elevation_ft
        if elev is None:
            return
        self.data.pressure_altitude = core.pressure_altitude(
            alt, elev, self.units.altimeter
        )
        self.data.density_altitude = core.density_altitude(alt, temp, elev, self.units)

    async def _post_update(self):
        if self.icao is None or self.raw is None:
            return
        self.data, self.units = parse(self.icao, self.raw, self.issued)
        if self._should_check_default:
            await self._pull_from_default()
        if self.data is None or self.units is None:
            return
        self._calculate_altitudes()
        self.translations = translate_metar(self.data, self.units)

    def _post_parse(self):
        if self.icao is None or self.raw is None:
            return
        self.data, self.units = parse(self.icao, self.raw, self.issued)
        if self.data is None or self.units is None:
            return
        self._calculate_altitudes()
        self.translations = translate_metar(self.data, self.units)

    @staticmethod
    def sanitize(report: str) -> str:
        """Sanitizes a METAR string"""
        return sanitize(report)[0]

    @property
    def summary(self) -> Optional[str]:
        """Condensed report summary created from translations"""
        if not self.translations:
            self.update()
        if self.translations is None:
            return None
        return summary.metar(self.translations)

    @property
    def speech(self) -> Optional[str]:
        """Report summary designed to be read by a text-to-speech program"""
        if not self.data:
            self.update()
        if self.data is None or self.units is None:
            return None
        return speech.metar(self.data, self.units)



"""
Remarks data translation handlers
"""

# stdlib
from typing import Dict, Optional
from avwx.structs import Number, PressureTendency, RemarksData


def temp_minmax(label: str, code: Number) -> str:
    """Translates a minimum or maximum temperature value"""
    return f"6-hour {label} temperature {code.value}°C"


def pressure_tendency(pressure: PressureTendency, unit: str = "mb") -> str:
    """Translates a pressure outlook value

    Ex: "50123" -> 12.3 mb: Increasing, then decreasing
    """
    value = f"{pressure.change} {unit} - {pressure.tendency}"
    return "3-hour pressure difference: +/- " + value


def precip(label: str, code: Number, unit: str = "in") -> str:
    """Translates a labelled precipitation value"""
    return f"Precipitation in the last {label}: {code.value} {unit}"


def sunshine_duration(code: Number, unit: str = "minutes") -> str:
    """Translates a sunlight duration value"""
    return f"Duration of sunlight: {code.value} {unit}"


def snow_depth(code: Number, unit: str = "in") -> str:
    """Translates a snow accumulation value"""
    return f"Snow accumulation: {code.value} {unit}"


def sea_level_pressure(code: Number) -> str:
    """Translates a sea level pressure value"""
    return f"Sea level pressure: {code.value} hPa"


def remarks_data(data: RemarksData) -> Dict[str, str]:
    """Extract translations from parsed remarks data"""
    ret = {}
    if data.temperature_decimal and data.dewpoint_decimal:
        temp, dew = data.temperature_decimal, data.dewpoint_decimal
        key = f"T{temp.repr}{dew.repr}"
        ret[key] = f"Temperature {temp.value}°C and dewpoint {dew.value}°C"
    if data.minimum_temperature_24 and data.maximum_temperature_24:
        minimum, maximum = data.minimum_temperature_24, data.maximum_temperature_24
        key = f"4{minimum.repr}{maximum.repr}"
        ret[key] = f"24-hour temperature: max {maximum.value} min {minimum.value}"
    if data.maximum_temperature_6:
        key = data.maximum_temperature_6.repr
        ret[key] = temp_minmax("maximum", data.maximum_temperature_6)
    if data.minimum_temperature_6:
        key = data.minimum_temperature_6.repr
        ret[key] = temp_minmax("minimum", data.minimum_temperature_6)
    if data.precip_36_hours:
        ret[data.precip_36_hours.repr] = precip("3/6 hours", data.precip_36_hours)
    if data.precip_24_hours:
        ret[data.precip_24_hours.repr] = precip("24 hours", data.precip_24_hours)
    if data.precip_hourly:
        ret[data.precip_hourly.repr] = precip("hour", data.precip_hourly)
    if data.pressure_tendency:
        ret[data.pressure_tendency.repr] = pressure_tendency(data.pressure_tendency)
    if data.sunshine_minutes:
        ret[data.sunshine_minutes.repr] = sunshine_duration(data.sunshine_minutes)
    if data.snow_depth:
        ret[data.snow_depth.repr] = snow_depth(data.snow_depth)
    if data.sea_level_pressure:
        ret[data.sea_level_pressure.repr] = sea_level_pressure(data.sea_level_pressure)
    return ret


def translate(raw: Optional[str], data: Optional[RemarksData]) -> Dict[str, str]:
    """Translates elements in the remarks string"""
    if not (raw and data):
        return {}
    ret = {}
    # Add static codes
    for code in data.codes:
        ret[code.repr] = code.value
    # Add features from the parsed remarks data
    ret.update(remarks_data(data))
    return ret



"""
avwx Package Setup
"""

from setuptools import find_namespace_packages, setup

VERSION = "1.6.16"

dependencies = [
    "geopy~=2.2",
    "httpx~=0.19",
    "python-dateutil~=2.8",
    "rapidfuzz~=1.6",
    "xmltodict~=0.12",
]

test_dependencies = ["pytest-asyncio>=0.15.1", "time-machine~=2.3"]

extras = {
    "scipy": ["scipy~=1.7"],
    "docs": ["mkdocs~=1.2", "mkdocs-material~=7.2", "mkdocs-minify-plugin~=0.4"],
    "tests": test_dependencies,
}

setup(
    name="avwx-engine",
    version=VERSION,
    description="Aviation weather report parsing library",
    url="https://github.com/avwx-rest/avwx-engine",
    author="Michael duPont",
    author_email="michael@dupont.dev",
    license="MIT",
    classifiers=[
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "License :: OSI Approved :: MIT License",
    ],
    python_requires=">= 3.7",
    install_requires=dependencies,
    packages=find_namespace_packages(include=["avwx*"]),
    package_data={"avwx.data": ["aircraft.json", "stations.json"]},
    tests_require=test_dependencies,
    extras_require=extras,
)



"""
Classes for retrieving raw report strings via managed files
"""

# pylint: disable=invalid-name,arguments-differ

# stdlib
import asyncio as aio
import atexit
import datetime as dt
import tempfile
from contextlib import suppress
from pathlib import Path
from socket import gaierror
from typing import Dict, Iterator, Optional, TextIO, Tuple

# library
import httpx

# module
from avwx.service.base import Service
from avwx.station import valid_station


_TEMP_DIR = tempfile.TemporaryDirectory()  # pylint: disable=consider-using-with
_TEMP = Path(_TEMP_DIR.name)


@atexit.register
def _cleanup():
    """Deletes temporary files and directory at Python exit"""
    _TEMP_DIR.cleanup()


class FileService(Service):
    """Service class for fetching reports via managed source files"""

    update_interval: dt.timedelta = dt.timedelta(minutes=10)
    _updating: bool = False

    @property
    def _file_stem(self) -> str:
        return f"{self.__class__.__name__}.{self.report_type}"

    @property
    def _file(self) -> Optional[Path]:
        """Path object of the managed data file"""
        for path in _TEMP.glob(self._file_stem + "*"):
            return path
        return None

    @property
    def last_updated(self) -> Optional[dt.datetime]:
        """When the file was last updated"""
        file = self._file
        if file is None:
            return None
        try:
            timestamp = int(file.name.split(".")[-2])
            return dt.datetime.fromtimestamp(timestamp, tz=dt.timezone.utc)
        except (AttributeError, ValueError):
            return None

    @property
    def is_outdated(self) -> bool:
        """If the file should be updated based on the update interval"""
        last = self.last_updated
        if last is None:
            return True
        now = dt.datetime.now(tz=dt.timezone.utc)
        return now > last + self.update_interval

    def _new_path(self) -> Path:
        now = dt.datetime.now(tz=dt.timezone.utc).timestamp()
        timestamp = str(now).split(".", maxsplit=1)[0]
        return _TEMP / f"{self._file_stem}.{timestamp}.txt"

    async def _wait_until_updated(self):
        while not self._updating:
            await aio.sleep(0.01)

    @property
    def _urls(self) -> Iterator[str]:
        raise NotImplementedError()

    def _extract(self, station: str, source: TextIO) -> Optional[str]:
        raise NotImplementedError()

    async def _update_file(self, timeout: int) -> bool:
        """Finds and saves the most recent file"""
        # Find the most recent file
        async with httpx.AsyncClient(timeout=timeout) as client:
            for url in self._urls:
                try:
                    resp = await client.get(url)
                    if resp.status_code == 200:
                        break
                except (httpx.ConnectTimeout, httpx.ReadTimeout):
                    return False
                except gaierror:
                    return False
            else:
                return False
        # Save successful file download
        new_path = self._new_path()
        with new_path.open("wb") as new_file:
            new_file.write(resp.content)
        return True

    async def update(self, wait: bool = False, timeout: int = 10) -> bool:
        """Update the stored file and returns success

        If wait, this will block if the file is already being updated
        """
        # Guard for other async calls
        if self._updating:
            if wait:
                await self._wait_until_updated()
                return True
            return False
        self._updating = True
        # Replace file
        old_path = self._file
        if not await self._update_file(timeout):
            self._updating = False
            return False
        if old_path:
            with suppress(FileNotFoundError):
                old_path.unlink()
        self._updating = False
        return True

    def fetch(
        self, station: str, wait: bool = True, timeout: int = 10, force: bool = False
    ) -> Optional[str]:
        """Fetch a report string from the source file

        If wait, this will block if the file is already being updated

        Can force the service to fetch a new file
        """
        return aio.run(self.async_fetch(station, wait, timeout, force))

    async def async_fetch(
        self, station: str, wait: bool = True, timeout: int = 10, force: bool = False
    ) -> Optional[str]:
        """Asynchronously fetch a report string from the source file

        If wait, this will block if the file is already being updated

        Can force the service to fetch a new file
        """
        valid_station(station)
        if wait and self._updating:
            self._wait_until_updated()
        if force or self.is_outdated:
            if not await self.update(wait, timeout):
                return None
        file = self._file
        if file is None:
            return None
        with file.open() as fin:
            report = self._extract(station, fin)
        return report


class NOAA_Forecast(FileService):
    """Subclass for extracting reports from NOAA FTP files"""

    def _index_target(self, station: str) -> Tuple[str, str]:
        raise NotImplementedError()

    def _extract(self, station: str, source: TextIO) -> Optional[str]:
        """Returns report pulled from the saved file"""
        start, end = self._index_target(station)
        txt = source.read()
        txt = txt[txt.find(start) :]
        txt = txt[: txt.find(end, 30)]
        lines = []
        for line in txt.split("\n"):
            if "CLIMO" not in line:
                line = line.strip()
            if not line:
                break
            lines.append(line)
        return "\n".join(lines) or None


class NOAA_NBM(NOAA_Forecast):
    """Requests forecast data from NOAA NBM FTP servers"""

    url = "https://nomads.ncep.noaa.gov/pub/data/nccf/com/blend/prod/blend.{}/{}/text/blend_{}tx.t{}z"
    _valid_types = ("nbh", "nbs", "nbe")

    @property
    def _urls(self) -> Iterator[str]:
        """Iterates through hourly updates no older than two days"""
        date = dt.datetime.now(tz=dt.timezone.utc)
        cutoff = date - dt.timedelta(days=1)
        while date > cutoff:
            timestamp = date.strftime(r"%Y%m%d")
            hour = str(date.hour).zfill(2)
            yield self.url.format(timestamp, hour, self.report_type, hour)
            date -= dt.timedelta(hours=1)

    def _index_target(self, station: str) -> Tuple[str, str]:
        return station + "   ", self.report_type.upper() + " GUIDANCE"


class NOAA_GFS(NOAA_Forecast):
    """Requests forecast data from NOAA GFS FTP servers"""

    url = "https://nomads.ncep.noaa.gov/pub/data/nccf/com/gfs/prod/gfsmos.{}/mdl_gfs{}.t{}z"
    _valid_types = ("mav", "mex")

    _cycles: Dict[str, Tuple[int, ...]] = {"mav": (0, 6, 12, 18), "mex": (0, 12)}

    @property
    def _urls(self) -> Iterator[str]:
        """Iterates through update cycles no older than two days"""
        now = dt.datetime.now(tz=dt.timezone.utc)
        date = dt.datetime.now(tz=dt.timezone.utc)
        cutoff = date - dt.timedelta(days=1)
        while date > cutoff:
            for cycle in reversed(self._cycles[self.report_type]):
                date = date.replace(hour=cycle)
                if date > now:
                    continue
                timestamp = date.strftime(r"%Y%m%d")
                hour = str(date.hour).zfill(2)
                yield self.url.format(timestamp, self.report_type, hour)
            date -= dt.timedelta(hours=1)

    def _index_target(self, station: str) -> Tuple[str, str]:
        return station + "   GFS", self.report_type.upper() + " GUIDANCE"



"""
Data load utilities
"""

# pylint: disable=missing-function-docstring

import json
from collections.abc import ItemsView, ValuesView
from pathlib import Path
from typing import Any, Callable, Dict, Iterable, Optional


class LazyLoad:
    """Lazy load a dictionary from the JSON data cache"""

    source: Path
    _data: Optional[Dict[str, Any]] = None

    def __init__(self, filename: str):
        self.source = Path(__file__).parent.joinpath("data", f"{filename}.json")

    def _load(self):
        self._data = json.load(self.source.open(encoding="utf8"))

    def _check(self):
        if self._data is None:
            self._load()

    @property
    def data(self) -> Dict[str, Any]:
        return self._data or {}

    def __getitem__(self, key: str) -> Any:
        self._check()
        return self.data[key]

    def __contains__(self, key: str) -> bool:
        self._check()
        return key in self.data

    def __len__(self) -> int:
        self._check()
        return len(self.data)

    def __iter__(self) -> Iterable[str]:
        self._check()
        for key in self.data:
            yield key

    def items(self) -> ItemsView:
        self._check()
        return self.data.items()

    def values(self) -> ValuesView:
        self._check()
        return self.data.values()


# LazyCalc lets us avoid the global keyword
class LazyCalc:
    """Delay data calculation until needed"""

    # pylint: disable=too-few-public-methods,missing-function-docstring

    _func: Callable
    _value: Optional[Any] = None

    def __init__(self, func: Callable):
        self._func = func  # type: ignore

    @property
    def value(self) -> Any:
        if self._value is None:
            self._value = self._func()
        return self._value
